body {
  color: white;
  background-color: #121212;
}

#header {
  background-color: #2f2f2f;
}

#main {
  color: white;
  background-color: #121212;
}

#sidebar {
  color: white;
  border-color: #2f2f2f !important;
  background-color: #121212;
}
